#!/usr/bin/env python3
# Gaia / Alien Prophets
import os
import json
import struct
import sys
import re
import copy
from decimal import Decimal
from pprint import pp

if len(sys.argv) < 3:
    sys.exit("Usage: compile-track.py <score.json> <bitstream.bs>\n")

def error(v, msg):
    if v != None:
        pp(v, indent = 4)
    
    sys.exit('ERROR: ' + msg)

def getValue(inp, key, typ, type_name, default):
    if not key in inp:
        if default == None:
            error(inp, 'Expected {0} value "{1}".'.format(type_name, key))
        else:
            return default

    if not type(inp[key]) is typ:
        error(inp, 'Expected the value "{0}" to be {1}, but got "{2}".'.format(key, type_name, type(inp[key]).__name__))
    
    return inp[key]

def getInt(inp, key, default):
    return getValue(inp, key, int, 'an integer', default)

def getFloat(inp, key, default):
    return getValue(inp, key, float, 'a float', default)

def getArray(inp, key, default):
    return getValue(inp, key, list, 'an array', default)

def getObject(inp, key, default):
    return getValue(inp, key, dict, 'an object', default)

def writeSampler(bs, v):
    t = 0

    if type(v) is Envelope:
        t = 1
    elif type(v) is Osc:
        t = 2
    elif type(v) is Mixer:
        t = 3
    elif type(v) is RefOsc:
        t = 6 # Osc with reference bit set
    
    bs.write(struct.pack('<H', t | ((v.idx & 0x1fff) << 3)))

class UIDMap:
    def __init__(self, type_name):
        self.type_name = type_name
        self.map = {}
    
    def validate(self, value, uid):
        if type(uid) != str:
            error(value, 'Expected UID to be a string but got "{0}".'.format(uid))

    def add(self, value, uid, idx):
        self.validate(value, uid)

        if uid in self.map:
            error(value, 'The {0} UID "{1}" is already used.'.format(self.type_name, uid))

        self.map[uid] = idx
    
    def resolve(self, value, uid):
        self.validate(value, uid)

        if uid not in self.map:
            error(value, 'The specified {0} UID "{1}" does not exist.'.format(self.type_name, uid))
        
        return self.map[uid]

class RefOsc:
    def __init__(self, parent, uid, idx):
        self.parent = parent
        self.uid = uid
        self.idx = idx

class Constant:
    def __init__(self, value, idx):
        self.value = value
        self.idx = idx

class Envelope:
    def __init__(self, score, value, idx):
        self.interp = value[0]
        self.idx = idx
        self.kp_start = None
        self.kp_end = None

        if self.interp != 'none' and self.interp != 'linear' and self.interp != 'cosine':
            error(value, 'The interpolation method of an envelope must be either "none", "linear" or "cosine".')
        
        env = value

        self.value = value[1:]

        if len(self.value) < 2:
            error(value, 'An envelope must contain at least two keypoints.')
        
        if len(self.value) > 256:
            error(value, 'An envelope can contain a maximum of 256 keypoints.')
        
        pvs = []

        for i in range(len(self.value)):
            v = self.value[i]
            pvs.append(self.parseKeypoint(score, env, i, v))
        
        self.value = pvs
    
    def parseKeypoint(self, score, env, idx, kp):
        kpl = len(kp)

        if kpl < 2 or kpl > 3:
            error(env, 'A keypoint must contain 2 elements (time, value) and - optionally - a sustain loop start / end tag.')
        
        if kp[0] < -16.0 or kp[0] > 49.535:
            error(env, 'An envelope keypoint must have a time between -16.0 and 49.535 seconds.')
        
        if kpl > 2:
            if type(kp[2]) != str or (kp[2] != 'start' and kp[2] != 'end'):
                error(env, 'When an envelope keypoint contains a third value, it must be either the string "start" or "end".')

            if kp[2] == 'start':
                if self.kp_start != None:
                    error(env, 'An envelope cannot specify more than one keypoint as sustain loop start.')
                
                self.kp_start = idx
            else:
                if self.kp_end != None:
                    error(env, 'An envelope cannot specify more than one keypoint as sustain loop end.')
                
                if self.kp_start == None:
                    error(env, 'An envelope cannot specify a keypoint as sustain loop end without a prior loop start.')
                
                self.kp_end = idx
        
        return [kp[0], score.parseValue(env, kp[1])]
    
    def write(self, bs):
        bs.write(struct.pack('B', len(self.value)))
        
        ff = 0 if self.interp == 'none' else 1 if self.interp == 'linear' else 2
        
        if self.kp_start != None:
            ff = ff | 4

        if self.kp_end != None:
            ff = ff | 8

        bs.write(struct.pack('B', ff))

        if self.kp_start != None:
            bs.write(struct.pack('B', self.kp_start))

        if self.kp_end != None:
            bs.write(struct.pack('B', self.kp_end))

        for v in self.value:
            bs.write(struct.pack('<H', int((v[0] + 16.0) * 1000.0)))
            writeSampler(bs, v[1])

def checkProps(tname, value, valid):
    for prop in value:
        if prop not in valid:
            error(value, 'The property "{0}" is not valid in a {1}.'.format(prop, tname))

class Filter:
    flt_types = { "lp6": 0, "lp12": 1, "lp18": 2, "lp24": 3, "hp6": 4, "hp12": 5, "hp18": 6, "hp24": 7, "bp12": 8, "bp24": 9, "notch12": 10, "notch18": 11, "ap12": 12, "ap24": 13, "phaser12": 14, "phaser24": 15 }
    valid_props = ['type', 'uid', 'gain', 'freq', 'res', 'mix']

    def __init__(self, score, value, osc_idx, idx):
        self.idx = idx

        checkProps('filter', value, Filter.valid_props)

        self.type = Filter.parseType(value)

        if 'uid' in value:
            score.flt_uids.add(value, value['uid'], [osc_idx, idx])

        for sampler in ['gain', 'freq', 'res', 'mix']:
            if sampler in value:
                setattr(self, sampler, score.parseValue(value, value[sampler]))
    
    def parseType(value):
        if not 'type' in value:
            return 3
        elif type(value['type']) is int:
            if value['type'] < 0 or value['type'] > 15:
                error(value, "When specified as a number, the filter type must be between 0 and 15.")

            return value['type']
        elif type(value['type']) is str:
            flt = value['type'].lower()

            if flt in Filter.flt_types:
                return Filter.flt_types[flt]
            
            error(value, 'Expected filter type (as string) to be one of "lp6/12/18/24", "hp6/12/18/24", "bp12/24" or "notch12/18", "ap12/24" or "phaser12/24" but got "{0}"'.format(flt))
        
        error(value, 'Expected the filter type to be either undeclared, an integer or a string')

    def write(self, bs):
        ff = self.type
        props = [['gain', 16], ['freq', 32], ['res', 64], ['mix', 128]]

        for a in props:
            if hasattr(self, a[0]):
                ff = ff | a[1]
        
        bs.write(struct.pack('B', ff))

        for a in props:
            if ff & a[1]:
                writeSampler(bs, getattr(self, a[0]))

class Osc:
    osc_types = { "filter": 0, "sampler": 1, "noise": 2, "sine": 3, "sawtooth": 4, "triangle": 5, "parabolic": 6, "square": 7 }
    valid_props = ['type', 'uid', 'lock', 'amp', 'freq', 'ofs', 'pha', 'pwm', 'freqmul', 'waveform', 'filters']

    def __init__(self, score, value):
        self.idx = len(score.oscs)
        score.oscs.append(self)

        checkProps('oscillator', value, Osc.valid_props)
        
        if 'uid' in value:
            score.osc_uids.add(value, value['uid'], self.idx)
        
        self.type = Osc.parseType(value)

        if self.type < 2:
            if not 'waveform' in value:
                error(value, 'When the osc type is "sampler" or "filter" a "waveform" value must be specified.')
            else:
                self.waveform = score.parseValue(value, value['waveform'])

        if 'lock' in value:
            self.lock = value['lock']
        else:
            self.lock = False if self.type > 0 else True

        for sampler in ['amp', 'freq', 'ofs', 'pha', 'pwm', 'freqmul']:
            if sampler in value:
                setattr(self, sampler, score.parseValue(value, value[sampler]))
        
        self.filters = []
        self.f_mode = 0 # Serial

        if 'filters' in value:
            if type(value['filters']) != list:
                error(value, 'The value of the "filters" property in an osc must be an array of filter objects.')
            
            if len(value['filters']) > 0 and type(value['filters'][0]) == str:
                if value['filters'][0] == 'par':
                    self.f_mode = 1
                
                value['filters'] = value['filters'][1:]
            
            for filter in value['filters']:
                self.filters.append(Filter(score, filter, self.idx, len(self.filters)))
            
            if len(self.filters) > 16:
                error(value, 'An oscillator can maximally have 16 filters.')
    
    def parseType(value):
        if not 'type' in value:
            error(value, 'An osc must specify a "type".')
        elif type(value['type']) is int:
            if value['type'] < 0 or value['type'] > 7:
                error(value, 'If an oscillator type is specified as an integer, it must be between 0 and 7.')

            return value['type']
        elif type(value['type']) is str:
            if value['type'] in Osc.osc_types:
                return Osc.osc_types[value['type']]
            
            error(value, 'Expected osc type (as string) to be one of "sine", "triangle", "sawtooth", "square", "parabolic", "sampler", "filter" or "noise", but got "{0}"'.format(value['type']))
        
        error(value, 'Expected the osc type to be either undeclared, an integer or a string.')

    def write(self, bs):
        samps = [['freq', 16], ['amp', 32], ['ofs', 64], ['pha', 128], ['pwm', 256], ['freqmul', 512]]

        ff = self.type

        if self.lock:
            ff = ff | 8
        
        for a in samps:
            if hasattr(self, a[0]):
                ff = ff | a[1]
        
        ff = ff | self.f_mode << 10
        ff = ff | (len(self.filters) << 11)

        bs.write(struct.pack('<H', ff))

        for a in samps:
            if ff & a[1]:
                writeSampler(bs, getattr(self, a[0]))

        if self.type < 2:
            writeSampler(bs, self.waveform)

        for f in self.filters:
            f.write(bs)

class Mixer:
    mixer_types = { 'add': 0, 'mul': 1, 'div': 2, 'cos': 3, 'min': 4, 'max': 5, 'map': 6, 'compress': 7, 'inst-time': 8, 'pat-time': 9, 'pitch': 10 }

    def __init__(self, score, value, idx):
        self.idx = idx
        score.mixers.append(self)
        self.elements = []

        if not value[0] in Mixer.mixer_types:
            p_idx = self.parseParam(value[0])

            if p_idx == None:
                error(value, 'A mixer must either specify its operation or an named parameter as a first string element. One of: "add", "mul", "div", "cos", "min", "max", "map", "compress", "inst-time", "pat-time", "pitch" or "param-0..43".')
            
            self.op = p_idx + 11
        else:
            self.op = Mixer.mixer_types[value[0]]
        
        for ei in range(1, len(value)):
            self.elements.append(score.parseValue(value, value[ei]))
        
        if self.op > 7 and len(self.elements) > 0:
            error(value, 'A mixer of this type cannot contain any elements.')

        if self.op < 8 and (len(self.elements) < 1 or len(self.elements) > 1023):
            error(value, 'A mixer of this type cannot be empty or contain more than 1023 nested elements.')
    
    def parseParam(self, s):
        if s.startswith('param-'):
            if len(s) > 6 and len(s) < 9:
                ss = s[6:8]

                if ss.isdigit():
                    idx = int(s[6:8])

                    if idx > -1 and idx < 44:
                        return idx
        
        return None
        
    def write(self, bs):
        bs.write(struct.pack('<H', self.op | (len(self.elements) << 6)))

        for v in self.elements:
            writeSampler(bs, v)

def checkUID(value):
    if 'uid' not in value:
        error(value, 'Expected an "uid" property.')

def checkProp(kind, value, valid):
    if 'prop' not in value:
        error(value, 'A propery ("prop") must be specified.')
    
    if value['prop'] not in valid:
        keys = list(valid.keys())
        error(value, 'Expected {0} property to be one of {1} or "{2}"'.format(kind, ', '.join(['"{0}"'.format(x) for x in keys[0:-1]]), keys[-1]))

def checkValue(kind, value):
    if not 'value' in value:
        error(value, 'A {0} command must specify a "value".'.format(kind))

class Command:
    osc_props = { "freq": 0, "amp": 1, "ofs": 2, "pha": 3, "pwm": 4, "freqmul": 5, "waveform": 6 }
    flt_props = { "gain": 0, "freq": 1, "res": 2, "mix": 3 }
    eff_props = { "vol": 0, "pan": 1, "detune": 2, "dry": 3, "delay": 4, "feedback": 5, "gain": 6, "freq": 7, "res": 8, "mix": 9 }

    def __init__(self, score, value):
        self.row = value[0]

        if self.row > 255:
            error(value, 'The row of a command must be between 0 and 255.')
        
        value = value[1]

        self.type = value['do']
        self.u = 0
        self.params = []

        if self.type == 'Note Off':
            self.type = 0
        elif self.type == 'Note On':
            self.type = 1
            
            pitch = value['pitch']

            if type(pitch) == str:
                pitch = Command.note_to_pitch(pitch)

                if pitch == None:
                    error(value, 'The pitch "{0}" could not be parsed.'.format(value['pitch']))
            elif type(pitch) == int:
                if pitch < -127 or pitch > 127:
                    error(value, 'When note on pitch is specified as an integer, it must be between -127 and 127.')
            else:
                    error(value, 'A note on pitch must be either a string or an integer.')

            if 'inst' not in value or type(value['inst']) != int:
                error(value, 'A note on command must have an integer "inst" property.')
            
            if value['inst'] < 0 or value['inst'] > len(score.instruments) - 1:
                error(value, 'A note on command used an invalid instrument index "{0}".'.format(value['inst']))
            
            self.params = [value['inst'], pitch]
        elif self.type == 'Osc Type':
            checkUID(value)
            self.type = 2
            self.u = Osc.parseType(value)
            self.params = [score.osc_uids.resolve(value, value['uid'])]
        elif self.type == 'Filter Type':
            checkUID(value)
            self.type = 3
            self.u = Filter.parseType(value)
            fref = score.flt_uids.resolve(value, value['uid'])
            self.params = [fref[0], fref[1]]
        elif self.type == 'Osc Prop':
            self.type = 4
            checkUID(value)
            checkProp('oscillator', value, Command.osc_props)
            checkValue('Osc Prop', value)
            self.u = Command.osc_props[value['prop']]
            self.params = [score.osc_uids.resolve(value, value['uid']), score.parseValue(value, value['value'])]
        elif self.type == 'Filter Prop':
            self.type = 5
            checkUID(value)
            checkProp('filter', value, Command.flt_props)
            checkValue('Filter Prop', value)
            self.u = Command.flt_props[value['prop']]
            fref = score.flt_uids.resolve(value, value['uid'])
            self.params = [fref[0], score.parseValue(value, value['value']), fref[1]]
        elif self.type == 'Track Prop':
            self.type = 6
            checkProp('track', value, Command.eff_props)
            checkValue('Track Prop', value)

            pidx = Command.eff_props[value['prop']]

            # Allow the user the specify a note value as the delay length and
            # have the compiler automatically calculate reciprocal pitch to make
            # Karplus-Strong synthesis more convenient.
            rpitch = None

            if pidx == 4 and type(value['value']) == str:
                rpitch = Command.note_to_pitch(value['value'])

                if rpitch != None:
                    rpitch = 440.0 * pow(2.0, (rpitch - 9.0) / 12.0) # Convert MIDI note to pitch
                    rpitch = score.insertConstant(1.0 / rpitch)

            self.params = [pidx, score.parseValue(value, value['value']) if rpitch == None else rpitch]

            if pidx > 3:
                self.u = self.parseTab(value)
        elif self.type == 'Delay Filter Type':
            self.type = 7
            self.u = self.parseTab(value)
            self.params = [Filter.parseType(value)]
        elif self.type == 'Halt':
            self.type = 8
        elif self.type == 'Off':
            self.type = 9
        elif self.type == 'Set Param':
            self.type = 10
            checkValue('Set Param', value)

            if 'index' in value and type(value['index']) == int and value['index'] >= 0 and value['index'] <= 43:
                self.params = [value['index'], score.parseValue(value, value['value'])]
            else:
                error(value, 'The "index" property of a "Set Param" command must be an integer between 0 and 43.')
        elif self.type == 'Set Tempo':
            self.type = 11
            checkValue('Set Tempo', value)
            self.params = [score.parseValue(value, value['value'])]
        else:
            error(value, 'Unsupported pattern command type "{0}".'.format(self.type))
        
    def parseTab(self, value):
        if not 'tab' in value:
            error(value, 'A track prop command of this type must specify a delay tab to which it applies.')
        
        if type(value['tab']) != int:
            error(value, 'The tab property must be an integer.')
        
        tab = int(value['tab'])

        if tab < 0 or tab > 15:
            error(value, 'The specified delay tab is invalid. It must be an integer between 0 and 15.')
        
        return tab

    note_keys = ['c-', 'c#', 'd-', 'd#', 'e-', 'f-', 'f#', 'g-', 'g#', 'a-', 'a#', 'b-']

    def note_to_pitch(note):
        if len(note) == 3:
            key = note[:2].lower()
            
            if key in Command.note_keys:
                key = Command.note_keys.index(key)
                
                if note[2].isdigit():
                    return key + ((int(note[2]) + 1) * 12) - 60
        
        return None

    def __repr__(self):
        return 'Cmd(type: {0}, u: {1}, pidx: {2})\n'.format(self.type, self.u, self.params[0])
    
    def write(self, bs):
        bs.write(struct.pack('B', self.row))
        bs.write(struct.pack('B', self.type | (self.u << 4)))

        if self.type == 1: # Note On
            bs.write(struct.pack('B', self.params[0]))
            bs.write(struct.pack('b', self.params[1]))
        elif self.type == 2 or self.type == 3: # Set Osc or Filter Type
            bs.write(struct.pack('<H', self.params[0]))

            if self.type == 3:
                bs.write(struct.pack('B', self.params[1]))
        elif self.type == 4 or self.type == 5: # Set Osc / Filter Property
            bs.write(struct.pack('<H', self.params[0]))

            if self.type == 5:
                bs.write(struct.pack('B', self.params[2]))
            
            writeSampler(bs, self.params[1])
        elif self.type == 6:
            bs.write(struct.pack('B', self.params[0]))
            writeSampler(bs, self.params[1])
        elif self.type == 7:
            bs.write(struct.pack('B', self.params[0]))
        elif self.type == 10:
            bs.write(struct.pack('B', self.params[0]))
            writeSampler(bs, self.params[1])
        elif self.type == 11:
            writeSampler(bs, self.params[0])

class CmdBuf:
    def __init__(self):
        pass

    @staticmethod
    def from_json(score, row_count, value):
        cb = CmdBuf()

        if 'track' not in value or type(value['track']) != int:
            error(value, 'A track object must have an integer "track" index property.')
        
        cb.track_idx = value['track']

        if cb.track_idx < 0 or cb.track_idx > 255:
            error(value, 'A pattern track must have an index between 0 and 255.')

        if cb.track_idx >= score.track_count:
            error(value, 'The track index "{0}" cannot be larger than or equal to the maximum number of tracks in the score ({1}).'.format(cb.track_idx, score.track_count))
        
        cb.inherit = 0xff
        
        if 'inherit' in value:
            if type(value['inherit']) != int:
                error(value, 'When the inherit property is specified for a track, it must be an integer value.')
            
            if value['inherit'] < -1 or value['inherit'] >= cb.track_idx:
                error(value, 'The inherit property of a track must be larger than -2 and less than the current track index ({0}).'.format(cb.track_idx))

            if value['inherit'] > -1:
                cb.inherit = value['inherit']

        for prop in ['replace', 'mute', 'solo']:
            setattr(cb, prop, prop in value and value[prop] == True)

        cb.recv = None if 'recv' not in value else value['recv']

        if cb.recv != None:
            if type(cb.recv[0]) == int:
                cb.recv = [cb.recv]
            
            if len(cb.recv) < 1 or len(cb.recv) > 128:
                error(value, 'A track recv list must have at least 1 and at most 128 elements.')
            
            for re in cb.recv:
                if len(re) != 2:
                    error(value, 'A track recv list must have exactly two elements: The track index to recieve from, and a mixing coefficient value.')
                
                if type(re[0]) != int:
                    error(value, 'The first element of a track recv element must be the integer index of a track.')
                
                if re[0] < 0 or re[0] >= score.track_count:
                    error(value, 'The index of the track to recieve from must be larger than 0 and less than the maximum track index.')
                
                re[1] = score.parseValue(value, re[1])
        else:
            cb.recv = []
        
        cb.commands = []

        last_row = 0

        if 'cmds' in value:
            for cmd in value['cmds']:
                c = Command(score, cmd)

                if c.row < 0 or c.row > row_count:
                    error(value, 'A track command must specify a row between 0 and {0}. Got {1}.'.format(row_count - 1, c.row))
                
                if c.row < last_row:
                    error(value, 'A track command cannot specify a row that is less than the row of a prior track command.')
                
                last_row = c.row
                cb.commands.append(c)
        
        if len(cb.commands) > 65535:
            error(value, 'A pattern track must contain at most 65535 commands.')

        return cb
    
    @staticmethod
    def from_cmdlist(cmds, tidx):
        cb = CmdBuf()

        cb.track_idx = tidx
        cb.inherit = 0xff
        cb.replace = False
        cb.mute = False
        cb.solo = False
        cb.recv = []
        cb.commands = cmds

        return cb

    def write(self, bs):
        if self.mute:
            return
        
        bs.write(struct.pack('B', self.track_idx))
        bs.write(struct.pack('B', self.inherit))

        flags = (len(self.recv) << 1) | (1 if self.replace else 0)

        bs.write(struct.pack('B', flags))

        for re in self.recv:
            bs.write(struct.pack('B', re[0]))
            writeSampler(bs, re[1])
        
        bs.write(struct.pack('<H', len(self.commands)))

        fp = bs.tell()

        for cmd in self.commands:
            cmd.write(bs)
        
        print('    Track {0}: {1} command(s), {2} bytes'.format(self.track_idx, len(self.commands), bs.tell() - fp))

class Pattern:
    def __init__(self, idx):
        self.idx = idx
    
    @staticmethod
    def from_json(score, value, idx):
        p = Pattern(idx)

        p.inherit = getInt(value, 'inherit', 0xffff)
        p.rows = getInt(value, 'rows', None)
        p.buffers = []
        p.act_buffers = 0

        if p.inherit == -1:
            p.inherit = 0xffff
        
        if p.inherit != 0xffff:
            if p.inherit >= idx or p.inherit < 0 or p.inherit > 0xffff:
                error(value, 'The pattern with index "{0}" cannot inherit from pattern "{1}".'.format(idx, p.inherit))
        
        if p.rows < 1 or p.rows > 256:
            error(value, 'A pattern must have between 1 and 256 rows, but "{0}" was specified for pattern "{1}".'.format(p.rows, idx))
        
        reuse_guard = []
        last_tidx = -1
        any_solo = False

        for track in value['tracks']:
            cb = CmdBuf.from_json(score, p.rows, track)

            if cb.track_idx in reuse_guard:
                error(value, 'Pattern {0} contains multiple tracks using index {1}.'.format(idx, cb.track_idx))
            
            if cb.track_idx <= last_tidx:
                error(value, 'Tracks must be specified in ascending track index order (track {0} cannot be specified after track {1}).'.format(cb.track_idx, last_tidx))
            
            reuse_guard.append(cb.track_idx)
            p.buffers.append(cb)

            if cb.solo:
                any_solo = True
        
        if len(p.buffers) > 256:
            error(value, 'A pattern cannot contain more than 256 tracks.')
        
        if any_solo:
            for buf in p.buffers:
                if not buf.solo:
                    buf.mute = True
        
        for buf in p.buffers:
            if not buf.mute:
                p.act_buffers += 1
        
        return p
    
    def write(self, bs):
        print('Pattern {0}'.format(self.idx))

        bs.write(struct.pack('<H', self.inherit))
        bs.write(struct.pack('B', self.rows - 1))
        bs.write(struct.pack('B', self.act_buffers))

        for buf in self.buffers:
            buf.write(bs)

def to22b(count):
    return [(count >> 16) & 0x3f, (count >> 8) & 0xff, count & 0xff]

class Score:
    def __init__(self, inp):
        self.track_count = getInt(inp, 'tracks', None)
        self.constants = []
        self.delay_params = []
        self.envelopes = []
        self.oscs = []
        self.mixers = []
        self.osc_uids = UIDMap('oscillator')
        self.flt_uids = UIDMap('filter')
        self.instruments = []
        self.patterns = []
        self.osc_refs = []
        self.fixed_consts = [Constant(1.0, 0), Constant(0.0, 1), Constant(1000.0, 2), Constant(0.5, 3), Constant(120.0, 4)]

        if self.track_count < 1 or self.track_count > 255:
            error(None, 'The specified number of tracks is outside the permissable range of 1-255.')

        if 'delay' in inp:
            delay = inp['delay']

            if type(delay) != list:
                error(delay, 'If a score "delay" property is specified, it must be an array of objects.')
            
            for d in delay:
                if type(d) != dict:
                    error(d, 'Each member of the score "delay" array must be an object with two keys: "track" (int) and "max-time" (float).')
                
                de = []

                if 'track' not in d or type(d['track']) != int or d['track'] < 0 or d['track'] >= self.track_count:
                    error(d, 'A delay parameter must specify a valid track index ("track") between 0 and {0}.'.format(self.track_count - 1))
                
                de.append(d['track'])

                if 'max-time' not in d or type(d['max-time']) != float or d['max-time'] * 44100 <= 2.0:
                    error(d, 'A delay parameter must specify a maximum delay time in seconds ("max-time"), which must be >= 1/22050 (0.00004536 seconds).')

                de.append(d['max-time'])

                for dp in self.delay_params:
                    if de[0] == dp[0]:
                        error(d, 'A delay parameter has already been specified for track "{0}".'.format(de[0]))
                
                self.delay_params.append(de)

        if not 'instruments' in inp or len(inp['instruments']) < 1:
            error(None, 'A score must define at least one instrument.')
        
        for inst in inp['instruments']:
            self.instruments.append(self.parseValue(inst, inst))
        
        if not 'order' in inp or type(inp['order']) != list or len(inp['order']) < 1 or len(inp['order']) > 65535:
            error(None, 'A valid score must contain an "order" array containing more than zero but less than 65536 integers.')
        
        self.order = inp['order']

        pidx = 0

        for pat in inp['patterns']:
            self.patterns.append(Pattern.from_json(self, pat, pidx))
            pidx += 1
        
        if pidx < 1:
            error(None, 'A valid score must contain at least one pattern.')
        
        for o in inp['order']:
            if type(o) != int or o < 0 or o >= pidx:
                error(inp['order'], 'The order array can only contain integers, which must all be valid pattern indexes.')
        
        if len(self.instruments) > 256:
            error(None, 'The total number of instruments in a score cannot exceed 256.')

        if len(self.constants) > 8187:
            error(None, 'The total number of constants in a score cannot exceed 8187.')

        if len(self.envelopes) > 8192:
            error(None, 'The total number of envelopes in a score cannot exceed 8192.')
        
        if len(self.oscs) > 8192:
            error(None, 'The total number of oscs in a score cannot exceed 8192.')

        if len(self.mixers) > 8192:
            error(None, 'The total number of mixers in a score cannot exceed 8192.')

        for ro in self.osc_refs:
            ro.idx = self.osc_uids.resolve(ro.parent, ro.uid)
        
        self.apply_scope(inp)
        self.optimize()
    
    def match_cmd(self, cmd, oc):
        if cmd.type != oc.type:
            return False
        
        if cmd.type == 2 and cmd.params[0] == oc.params[0]:
            return True
        elif cmd.type == 3 and cmd.params[0] == oc.params[1] and cmd.params[1] == oc.params[1]:
            return True
        elif cmd.type == 4 and cmd.u == oc.u and cmd.params[0] == oc.params[0]:
            return True
        elif cmd.type == 5 and cmd.u == oc.u and cmd.params[0] == oc.params[1] and cmd.params[2] == oc.params[2]:
            return True
        elif cmd.type == 6 and cmd.params[0] == oc.params[0]:
            if cmd.params[0] < 4 or cmd.u == oc.u:
                return True
        elif cmd.type == 7 and cmd.u == oc.u:
            return True
        elif cmd.type == 8 or cmd.type == 9:
            return True
        elif cmd.type == 10 and cmd.params[0] == oc.params[0]:
            return True
        
        return False
    
    def search_cmds(self, cmd, ntd):
        for c in ntd:
            if self.match_cmd(cmd, c):
                return True
        
        return False
    
    def apply_scope(self, inp):
        if 'scope' not in inp:
            return
        
        if type(inp['scope']) is not dict:
            error(None, 'When the score specifies a "scope" property its type must be an object.')
        
        scp = inp['scope']

        for prop in ['start', 'count']:
            if prop not in scp or type(scp[prop]) is not int:
                error(scp, 'When a scope object is specified, it must contain a "{0}" property the value of which must be an integer.'.format(prop))
        
        start = scp['start']
        count = scp['count']

        if start < 0 or start >= len(self.order):
            error(scp, 'The scope specified an invalid "start" value, which must be between 0 and {0}.'.format(len(self.order) - 1))
        
        if count < 1 or start + count > len(self.order):
            error(scp, 'The scope specified an invalid "count" value, which must be between 1 and {0}'.format(len(self.order) - start))

        tdata = [[] for x in range(self.track_count)]

        # Collate preceeding track data
        for oi in range(start):
            for cb in self.patterns[self.order[oi]].buffers:
                tdata[cb.track_idx].extend(cb.commands)
        
        # Optimize the resulting collated command sets
        ntdata = []
        stcmd = None

        for td in tdata:
            ntd = []

            for cmd in td:
                if cmd.type == 11: # Set tempo
                    stcmd = cmd
            
            for cmd in reversed(td):
                if cmd.type < 2 or cmd.type == 11: # Drop Note on / off and Set tempo commands
                    continue
                
                if self.search_cmds(cmd, ntd):
                    continue
                
                nc = copy.copy(cmd)
                nc.row = 0
                ntd.insert(0, nc)
            
            ntdata.append(ntd)
        
        if stcmd != None:
            nst = copy.copy(stcmd)
            nst.row = 0
            ntdata[0].insert(0, nst)
        
        # Inject reduced command set in start pattern data
        tp = Pattern(len(self.patterns))
        tp.inherit = 0xffff
        tp.rows = 1
        tp.buffers = []
        tp.act_buffers = 0

        for i in range(self.track_count):
            if len(ntdata[i]) < 1:
                continue
            
            print('* Inserting track {0}: {1}'.format(i, ntdata[i]))
            tp.buffers.append(CmdBuf.from_cmdlist(ntdata[i], i))
            tp.act_buffers += 1
        
        self.patterns.append(tp)

        # Apply scrope constraints to order
        self.order = self.order[start:start+count]
        self.order.insert(0, tp.idx)
        
        print(self.order)

    def optimize(self):
        # Sort constants by sign, exponent, digits to make them generally compress better
        def c_sort(c):
            cd = Decimal(c.value).as_tuple()
            return (cd[0], cd[2], cd[1])
        
        self.constants.sort(key = c_sort)

        idx = len(self.fixed_consts)

        for e in self.constants:
            e.idx = idx
            idx += 1

    def insertConstant(self, value):
        for c in self.fixed_consts:
            if c.value == value:
                return c
        
        for c in self.constants:
            if c.value == value:
                return c
        
        c = Constant(value, -1)
        self.constants.append(c)

        return c
    
    def insertEnvelope(self, value):
        ne = Envelope(self, value, -1)

        for e in self.envelopes:
            if len(e.value) != len(ne.value):
                continue
            
            match = e.interp == ne.interp

            if match:
                for vi in range(len(e.value)):
                    if type(e.value[vi][1]) != Constant or type(ne.value[vi][1]) != Constant:
                        match = False
                        break

                    if e.value[vi][0] != ne.value[vi][0] or e.value[vi][1].value != ne.value[vi][1].value:
                        match = False
                        break
            
            if match:
                return e
        
        ne.idx = len(self.envelopes)
        self.envelopes.append(ne)
        return ne
    
    def parseValue(self, parent, value):
        av = None

        if type(value) is float or type(value) is int:
            av = self.insertConstant(float(value))
        elif type(value) is list:
            if len(value) < 1:
                error(parent, 'Arrays are expected to contain at least one element.')
            
            if not type(value[0]) is str:
                error(parent, 'Any array must contain a string as its first element. Specify either "none", "linear" or "cosine" to denote an envelope or one of the mixer types to denote a mixer.')
            
            if value[0] == 'none' or value[0] == 'linear' or value[0] == 'cosine':
                av = self.insertEnvelope(value)
            else: # Mixer
                av = Mixer(self, value, len(self.mixers))
        elif type(value) is dict:
            av = Osc(self, value)
        elif type(value) is str:
            av = RefOsc(parent, value, -1)
            self.osc_refs.append(av)

        if av is None:
            error(parent, 'Expected a number, envelope, mixer, oscillator or oscillator reference, but got "{0}".'.format(value))

        return av
    
    def printSize(self, bs, name, col, fp):
        print('Number of {0}: {1} ({2} bytes)'.format(name, len(col), bs.tell() - fp))
        return bs.tell()

    def write(self, filename):
        with open(filename, 'wb') as bs:
            fp = bs.tell()

            bs.write(struct.pack('B', self.track_count))
            bs.write(struct.pack('B', len(self.delay_params)))
            bs.write(struct.pack('B', len(self.instruments)))
            bs.write(struct.pack('<H', len(self.constants)))
            bs.write(struct.pack('<H', len(self.envelopes)))
            bs.write(struct.pack('<H', len(self.oscs)))
            bs.write(struct.pack('<H', len(self.mixers)))
            bs.write(struct.pack('<H', len(self.order)))
            bs.write(struct.pack('<H', len(self.patterns)))

            print('Header ({0} bytes)'.format(bs.tell() - fp))
            fp = bs.tell()

            for c in self.constants:
                bs.write(struct.pack('<f', c.value))
            
            fp = self.printSize(bs, 'constants', self.constants, fp)
            
            for dp in self.delay_params:
                bs.write(struct.pack('B', dp[0]))
                bs.write(struct.pack('<f', dp[1]))
            
            fp = self.printSize(bs, 'delay params', self.delay_params, fp)

            for e in self.envelopes:
                e.write(bs)
            
            fp = self.printSize(bs, 'envelopes', self.envelopes, fp)

            for o in self.oscs:
                o.write(bs)

            fp = self.printSize(bs, 'oscs', self.oscs, fp)

            for m in self.mixers:
                m.write(bs)
            
            fp = self.printSize(bs, 'mixers', self.mixers, fp)

            for i in self.instruments:
                writeSampler(bs, i)
            
            fp = self.printSize(bs, 'instruments', self.instruments, fp)
            
            for o in self.order:
                bs.write(struct.pack('<H', o))
            
            print('Number of osc uids: {0}'.format(len(self.osc_uids.map)))
            print('Number of filter uids: {0}'.format(len(self.flt_uids.map)))

            fp = self.printSize(bs, 'order indices', self.order, fp)

            for p in self.patterns:
                p.write(bs)
            
            fp = self.printSize(bs, 'patterns', self.patterns, fp)
            print("Final output size: {0} bytes".format(fp))

def loadJSONWithComments(filename):
    cre = r'(/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/)|(//.*)'

    with open(sys.argv[1], 'rt') as inp_f:
        return re.sub(cre, '', inp_f.read())

inp = json.loads(loadJSONWithComments(sys.argv[1]))
Score(inp).write(sys.argv[2])
os.system('xz -z -9 -e -f -k {0}'.format(sys.argv[2]))
print('Estimated compressed output size: {0} bytes'.format(os.path.getsize(sys.argv[2] + '.xz')))
os.system('rm {0}.xz'.format(sys.argv[2]))