#ifdef NOMUSIC
#pragma GCC diagnostic ignored "-Wunused-variable"
#pragma GCC diagnostic ignored "-Wunused-const-variable"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mother.h"

static float Sample(Score* sc, Sampler* s, float rt);
static void Trigger(Score* sc, Sampler* s, unsigned int cmd, float param);

static const float PI   = 3.14159265f;
static const float TAU  = 6.28318531f;
static const float N3DB = 0.70710678f;

static unsigned long       pcg_sta = 0x4d595df4d0f33173;
static const unsigned long pcg_mul = 6364136223846793005u;
static const unsigned long pcg_inc = 1442695040888963407u;

static const signed char pole_weights[16][5] = 
{ 
    {0, -1,  0,   0,  0}, // LP6
    {0,  0,  1,   0,  0}, // LP12
    {0,  0,  0,  -1,  0}, // LP18
    {0,  0,  0,   0,  1}, // LP24
    {1, -1,  0,   0,  0}, // HP6
    {1, -2,  1,   0,  0}, // HP12
    {1, -3,  3,  -1,  0}, // HP18
    {1, -4,  6,  -4,  1}, // HP24
    {0, -1,  1,   0,  0}, // BP12
    {0,  0,  1,  -2,  1}, // BP24
    {1, -2,  2,   0,  0}, // NOTCH12
    {1, -3,  3,  -2,  0}, // NOTCH18
    {1, -4,  4,   0,  0}, // AP12
    {1, -8, 24, -32, 16}, // AP24
    {1, -2,  2,   0,  0}, // PHASER12
    {1, -4, 12, -16,  8}  // PHASER24
};

static unsigned int pcg32(void)
{
	unsigned long x = pcg_sta;
	const unsigned int c = (unsigned int)(x >> 59);

	pcg_sta = x * pcg_mul + pcg_inc;
	x ^= x >> 18;

	const unsigned int y = (unsigned int)(x >> 27);

	return (unsigned int)(y >> c | y << (-c & 31));
}

static float SgnMod(const float v, const float l)
{
    const float fm = v - ((float)(int)(v / l)) * l;

    return v >= 0.0f ? fm : l + fm;
}

static void Trigger(Score* sc, Sampler* s, unsigned int cmd, float param)
{
    if(s->type == ST_CONST)
        return;
    
    if(s->type == ST_OSC)
    {
        if(s->is_ref)
            return;
        
        Osc* osc = sc->tracks[sc->tidx].oscs + s->index;

        if(cmd == TC_NOTE_ON)
            osc->bf = osc->lock == 0 ? param : 0.0f;
        
        Trigger(sc, &osc->freq, cmd, param);
        Trigger(sc, &osc->amp, cmd, param);
        Trigger(sc, &osc->offset, cmd, param);
        Trigger(sc, &osc->phase, cmd, param);
        Trigger(sc, &osc->pwm, cmd, param);

        if(osc->type < OT_NOISE)
            Trigger(sc, &osc->waveform, cmd, param);
    }
    else if(s->type == ST_ENV)
    {
        Envelope* e = &sc->envelopes[s->index];

        if(cmd == TC_NOTE_ON)
            e->rel_time = -1.0f;
        else
            e->rel_time = param - e->act_time;

        for(unsigned int i = 0; i < e->size; i++)
            Trigger(sc, &e->values[i], cmd, param);
    }
    else // if(s->type == ST_MIX)
    {
        Mixer* om = &sc->mixers[s->index];

        for(unsigned int i = 0; i < om->count; i++)
            Trigger(sc, &om->elements[i], cmd, param);
    }
}

static float RunFilter(Score* sc, Filter* f, float rt, float v)
{
    const float freq = Sample(sc, &f->freq, rt),
                res = Sample(sc, &f->res, rt),
                gain = Sample(sc, &f->gain, rt),
                dw_mix = Sample(sc, &f->dw_mix, rt);

    if(f->freq_v != freq || f->res_v != res)
    {
        f->freq_v = freq;
        f->res_v = res;

        const float w = TAU * freq / 44100.0f, w2 = w * w, w3 = w2 * w, w4 = w3 * w;
        
        f->g = 0.9892f * w - 0.4342f * w2 + 0.1381f * w3 - 0.0202f * w4;
        f->gr = res * (1.0029f + 0.0526f * w - 0.0926f * w2 + 0.0218f * w3) * 4.0f;

        for(unsigned int i = 0; i < 4; i++)
            f->rcf[i].g = f->g;
    }

	const float fb = v - ((f->d - (v * 0.5f)) * f->gr);
    const signed char* pw = &pole_weights[f->type][0];
    const float fb2 = fb * fb, fb3 = fb2 * fb, fb4 = fb3 * fb;

	float s = 0.0f;
    float fo = (fb + (0.43157974f * fb3)) / (1.0f + (0.76443945f * fb2) + (0.05831938f * fb4));

	for(unsigned int i = 0; i < 4; i++)
    {
        RCFilter* rcf = f->rcf + i;
        const float o = (((0.76923076f * fo) + (0.23076923f * rcf->lv) - rcf->lo) * rcf->g) + rcf->lo;

        s += fo * (float)pw[i];
        rcf->lv = fo;
        rcf->lo = fo = o;
	}

	f->d = fo;
	s += fo * (float)pw[4];

	return ((v * (1.0f - dw_mix)) + (s * dw_mix)) * gain;
}

static float SampleOsc(Score* sc, Osc* osc, float rt)
{
    float v = 0.0f;

    if(osc->type == OT_NOISE)
    {
        v = (((float)pcg32() / (float)0xffffffff) * 2.0f) - 1.0f;
    }
    else
    {
        const float p_ofs = Sample(sc, &osc->phase, rt);
        
        if(osc->type == OT_FILTER)
        {
            v = Sample(sc, &osc->waveform, rt + p_ofs);
        }
        else
        {
            float pwm = Sample(sc, &osc->pwm, rt);
            float p = SgnMod(osc->p + p_ofs, 1.0f);

            if(pwm != 0.5f)
            {
                pwm = pwm < 0.00001f ? 0.00001f : pwm > 0.99999f ? 0.99999f : pwm;
                p = p <= pwm ? (p / pwm) * 0.5f : (((p - pwm) / (1.0f - pwm)) * 0.5f) + 0.5f;
            }

            if(osc->type == OT_SAMPLER)
            {
                v = Sample(sc, &osc->waveform, p);
            }
            else if(osc->type == OT_SINE)
            {
                v = sinf(p * TAU);
            }
            else
            {
                const float t = (p - 0.5f) * 2.0f;

                if(osc->type == OT_SAWTOOTH)
                {
                    v = t;
                }
                else if(osc->type == OT_TRIANGLE)
                {
                    v = t * 2.0f;
                    v = v < 0.0f ? v + 1.0f : 1.0f - v;
                }
                else if(osc->type == OT_PARABOLIC)
                {
                    v = ((t * t) * 2.0f) - 1.0f;
                }
                else // if(osc->type == OT_SQUARE)
                {
                    v = t > 0.0f ? 1.0f : -1.0f;
                }
            }

            osc->p += ((osc->bf + sc->detune + Sample(sc, &osc->freq, rt)) * Sample(sc, &osc->freqmul, rt)) / 44100.0f;
            osc->p = SgnMod(osc->p, 1.0f);
        }
    }

    v *= Sample(sc, &osc->amp, rt);
    v += Sample(sc, &osc->offset, rt);
    
    if(osc->f_mode == 0)
    {
        for(unsigned int i = 0; i < osc->f_count; i++)
            v = RunFilter(sc, &osc->filters[i], rt, v);
    }
    else
    {
        float s = 0.0f;

        for(unsigned int i = 0; i < osc->f_count; i++)
            s += RunFilter(sc, &osc->filters[i], rt, v);
        
        v = s;
    }
    
    osc->lv = v;

    return v;
}

static float SampleEnv(Score* sc, Envelope* e, float rt)
{
    e->act_time = rt;

    if(e->sust_start > -1)
    {
        const float ss = e->tcs[e->sust_start];

        if(rt >= ss)
        {
            if(e->rel_time == -1.0f)
            {
                if(e->sust_end > -1)
                    e->act_time = SgnMod(rt - ss, e->tcs[e->sust_end] - ss) + ss;
                else
                    e->act_time = ss;
            }
            else
                e->act_time = rt - e->rel_time;
        }
    }

    while(e->act_time >= e->tcs[e->idx])
    {
        if(e->idx == e->size - 1)
            return Sample(sc, &e->values[e->idx], e->act_time);
        
        e->idx++;
    }

    while(e->act_time < e->tcs[e->idx-1])
    {
        if(e->idx == 1)
            return Sample(sc, &e->values[0], e->act_time);
        
        e->idx--;
    }
    
    const float lv = Sample(sc, &e->values[e->idx-1], e->act_time),
                lt = e->tcs[e->idx-1];

    if(e->intp != ENV_NONE)
    {
        const float t = (e->act_time - lt) / (e->tcs[e->idx] - lt);
        const float t2 = e->intp == ENV_COSINE ? (1.0f - cosf(t * PI)) / 2.0f : t;
        
        return ((1.0f - t2) * lv) + (t2 * Sample(sc, &e->values[e->idx], e->act_time));
    }

    return lv;
}

static float Sample(Score* sc, Sampler* s, float rt)
{
    if(s->type == ST_CONST)
        return sc->constants[s->index];
    else if(s->type == ST_ENV)
        return SampleEnv(sc, &sc->envelopes[s->index], rt);
    else if(s->type == ST_OSC)
    {
        Osc* osc = &sc->tracks[sc->tidx].oscs[s->index];

        if(s->is_ref)
            return osc->lv;
        
        return SampleOsc(sc, osc, rt);
    }

    Mixer* m = &sc->mixers[s->index];

    if(m->op > OMOP_COMPRESS)
    {
        if(m->op == OMOP_INST_TIME)
            return sc->abst - sc->tracks[sc->tidx].st;
        else if(m->op == OMOP_PAT_TIME)
            return sc->pt;
        else if(m->op == OMOP_PITCH)
            return sc->tracks[sc->tidx].pitch + sc->detune;
        
        return Sample(sc, &sc->tracks[sc->tidx].params[m->op-11], sc->pt);
    }

    float v = Sample(sc, &m->elements[0], rt);

    for(unsigned int i = 1; i < m->count; i++)
    {
        if(m->op < OMOP_MAP)
        {
            const float ov = Sample(sc, &m->elements[i], rt);

            if(m->op == OMOP_ADD)
                v += ov;
            else if(m->op == OMOP_MUL)
                v *= ov;
            else if(m->op == OMOP_DIV)
                v /= (ov < 0.00001f ? 0.00001f : ov);
            else if(m->op == OMOP_COS)
                v *= cosf(v * ov);
            else
            {
                const float as = fabsf(ov),
                            av = fabsf(v),
                            sv = as * (v < 0.0f ? -1.0f : 1.0f);
                
                const int sm = av < as;

                v = m->op == OMOP_MIN ? (sm ? v : sv) : (sm ? sv : v);
            }
        }
        else
        {
            const float mf = Sample(sc, &m->elements[i], v);

            if(m->op == OMOP_MAP)
                v = mf;
            else
                v *= mf;
        }
    }

    return v;
}

static unsigned char GetByte(const unsigned char** bs)
{
    unsigned char t = (*bs)[0];

    (*bs)++;
    return t;
}

static unsigned short GetShort(const unsigned char** bs)
{
    unsigned short t = ((unsigned short*)(*bs))[0];

    (*bs)+=2;
    return t;
}

static float GetFloat(const unsigned char** bs)
{
    float t = ((float*)(*bs))[0];

    (*bs)+=4;
    return t;
}

static void GetSampler(const unsigned char** bs, Sampler* s, unsigned int present, unsigned int def)
{
    if(!present)
    {
        s->type = ST_CONST;
        s->index = def;
        return;
    }

    unsigned int t = GetShort(bs);

    s->type = t & 3;
    s->is_ref = t & 4;
    s->index = (t >> 3) & 0x1fff;
}

static void InitFilter(Filter* f, unsigned int type)
{
    f->type = type;
    memset(&f->rcf[0], 0, sizeof(float) * 17);
}

static void Init(Score* s, const unsigned char** bs)
{
    s->track_count = GetByte(bs);
    s->dp_count = GetByte(bs);
    s->inst_count = GetByte(bs);
    s->const_count = GetShort(bs);
    s->env_count = GetShort(bs);
    s->osc_count = GetShort(bs);
    s->mix_count = GetShort(bs);
    s->order_count = GetShort(bs);
    s->pattern_count = GetShort(bs);
    s->constants = malloc((s->const_count + EC_COUNT) * sizeof(float));
    s->envelopes = malloc(s->env_count * sizeof(Envelope));
    s->oscs = malloc(s->osc_count * sizeof(Osc));
    s->mixers = malloc(s->mix_count * sizeof(Mixer));
    s->instruments = malloc(s->inst_count * sizeof(Sampler));
    s->order = malloc(s->order_count * sizeof(unsigned int));
    s->patterns = malloc(s->pattern_count * sizeof(Pattern));
    s->tracks = malloc(s->track_count * sizeof(Track));
    s->t_samps = malloc(s->track_count * sizeof(float));

    // Add the default values
    s->constants[EC_1] = 1.0f;
    s->constants[EC_0] = 0.0f;
    s->constants[EC_1000] = 1000.0f;
    s->constants[EC_05] = 0.5f;
    s->constants[EC_120] = 120.0f;

    for(unsigned int i = EC_COUNT; i < s->const_count + EC_COUNT; i++)
        s->constants[i] = GetFloat(bs);

    GetSampler(bs, &s->tempo, 0, EC_120);

    DelayParam* dps = malloc(s->dp_count * sizeof(DelayParam));

    for(unsigned int i = 0; i < s->dp_count; i++)
    {
        dps[i].tidx = GetByte(bs);
        dps[i].duration = GetFloat(bs);
    }

    for(unsigned int i = 0; i < s->env_count; i++)
    {
        Envelope* e = &s->envelopes[i];

        e->size = GetByte(bs);

        unsigned int ff = GetByte(bs);

        e->intp = ff & 3;
        e->idx = 1;
        e->sust_start = (ff & 4) ? GetByte(bs) : -1;
        e->sust_end = (ff & 8) ? GetByte(bs) : -1;
        e->rel_time = -1.0f;

        e->tcs = malloc(e->size * sizeof(float));
        e->values = malloc(e->size * sizeof(Sampler));

        for(unsigned int j = 0; j < e->size; j++)
        {
            e->tcs[j] = (GetShort(bs) / 1000.0f) - 16.0f;
            GetSampler(bs, &e->values[j], 1, 0);
        }
    }

    for(unsigned int i = 0; i < s->osc_count; i++)
    {
        Osc* osc = &s->oscs[i];
        unsigned int ff = GetShort(bs);

        osc->type = ff & 7;
        osc->lock = (ff & 8) >> 3;
        osc->f_count = ff >> 11;
        osc->f_mode = ff & 1024;
        osc->bf = osc->p = osc->lv = 0.0f;

        GetSampler(bs, &osc->freq, ff & 16, EC_0);
        GetSampler(bs, &osc->amp, ff & 32, EC_1);
        GetSampler(bs, &osc->offset, ff & 64, EC_0);
        GetSampler(bs, &osc->phase, ff & 128, EC_0);
        GetSampler(bs, &osc->pwm, ff & 256, EC_05);
        GetSampler(bs, &osc->freqmul, ff & 512, EC_1);
        
        if(osc->type < OT_NOISE)
            GetSampler(bs, &osc->waveform, 1, 0);

        if(osc->f_count > 0)
        {
            osc->filters = malloc(osc->f_count * sizeof(Filter));

            for(unsigned int j = 0; j < osc->f_count; j++)
            {
                Filter* f = &osc->filters[j];
                unsigned int flff = GetByte(bs);

                InitFilter(f, flff & 15);
                GetSampler(bs, &f->gain, flff & 16, EC_1);
                GetSampler(bs, &f->freq, flff & 32, EC_1000);
                GetSampler(bs, &f->res, flff & 64, EC_05);
                GetSampler(bs, &f->dw_mix, flff & 128, EC_1);
            }
        }
    }

    for(unsigned int i = 0; i < s->mix_count; i++)
    {
        Mixer* m = &s->mixers[i];
        unsigned short ff = GetShort(bs);
        
        m->op = ff & 0x3f;
        m->count = ff >> 6;
        m->elements = malloc(m->count * sizeof(Sampler));

        for(unsigned int j = 0; j < m->count; j++)
            GetSampler(bs, &m->elements[j], 1, 0);
    }

    for(unsigned int i = 0; i < s->inst_count; i++)
        GetSampler(bs, &s->instruments[i], 1, 0);

    for(unsigned int i = 0; i < s->order_count; i++)
        s->order[i] = GetShort(bs);
    
    for(unsigned int i = 0; i < s->track_count; i++)
    {
        Track* t = &s->tracks[i];

        memset(t, 0, sizeof(Track));

        for(unsigned int j = 0; j < s->dp_count; j++)
        {
            if(dps[j].tidx == (int)i)
            {
                t->d_size = (int)(dps[j].duration * 44100.0f);
                t->d_buf = malloc(t->d_size * sizeof(float));
                memset(t->d_buf, 0, t->d_size * sizeof(float));
                break;
            }
        }

        GetSampler(bs, &t->vol, 0, EC_1);
        GetSampler(bs, &t->pan, 0, EC_05);
        GetSampler(bs, &t->dry, 0, EC_1);
        GetSampler(bs, &t->detune, 0, EC_0);

        for(int j = 0; j < 44; j++)
            GetSampler(bs, &t->params[j], 0, EC_0);

        t->active_inst = -1;

        t->oscs = malloc(s->osc_count * sizeof(Osc));

        for(int j = 0; j < 16; j++)
        {
            t->filter[j] = malloc(sizeof(Filter));
            InitFilter(t->filter[j], FT_LP24);
            GetSampler(bs, &t->delay[j], 0, EC_0);
            GetSampler(bs, &t->feedback[j], 0, EC_1);
            GetSampler(bs, &t->filter[j]->gain, 0, EC_1);
            GetSampler(bs, &t->filter[j]->freq, 0, EC_1000);
            GetSampler(bs, &t->filter[j]->res, 0, EC_0);
            GetSampler(bs, &t->filter[j]->dw_mix, 0, EC_1);
        }

        memcpy(t->oscs, s->oscs, s->osc_count * sizeof(Osc));

        for(unsigned int j = 0; j < s->osc_count; j++)
        {
            Osc* to = &t->oscs[j];

            if(to->f_count)
            {
                to->filters = malloc(to->f_count * sizeof(Filter));
                memcpy(to->filters, s->oscs[j].filters, to->f_count * sizeof(Filter));
            }
            else
            {
                to->filters = NULL;
            }
        }
    }

    for(unsigned int i = 0; i < s->pattern_count; i++)
    {
        Pattern* p = &s->patterns[i];

        p->inherits = GetShort(bs);
        p->row_count = GetByte(bs) + 1;
        p->tracks = malloc(s->track_count * sizeof(CmdBuf*));
        p->row_frames = malloc(p->row_count * sizeof(unsigned int));

        unsigned int track_count = GetByte(bs);

        CmdBuf* tracks = malloc(track_count * sizeof(CmdBuf));

        for(unsigned int j = 0; j < s->track_count; j++)
            p->tracks[j] = NULL;

        for(unsigned int j = 0; j < track_count; j++)
        {
            CmdBuf* t = &tracks[j];

            t->track_idx = GetByte(bs);
            t->inherits = GetByte(bs);
            
            unsigned int flags = GetByte(bs);

            t->replace = flags & 1;
            t->r_count = flags >> 1;

            if(t->r_count > 0)
            {
                t->r_elems = malloc(t->r_count * sizeof(RecvElem));

                for(unsigned int k = 0; k < t->r_count; k++)
                {
                    t->r_elems[k].tidx = GetByte(bs);
                    GetSampler(bs, &t->r_elems[k].value, 1, 0);
                }
            }

            t->cmd_count = GetShort(bs);
            t->commands = malloc(t->cmd_count * sizeof(Command));

            for(unsigned int c = 0; c < t->cmd_count; c++)
            {
                Command* cmd = &t->commands[c];

                cmd->row = GetByte(bs);
                cmd->type = GetByte(bs);

                int at = cmd->type & 0xf;

                if(at == CMD_NOTE_ON)
                {
                    cmd->inst = GetByte(bs);
                    cmd->pitch = (signed char)GetByte(bs);
                }
                else if(at >= CMD_SET_OSC_TYPE && at <= CMD_SET_FLT_PROP)
                {
                    cmd->osc = &s->tracks[t->track_idx].oscs[GetShort(bs)];

                    if(at == CMD_SET_FLT_TYPE || at == CMD_SET_FLT_PROP)
                        cmd->filter = &cmd->osc->filters[GetByte(bs)];
                    
                    if(at == CMD_SET_OSC_PROP || at == CMD_SET_FLT_PROP)
                        GetSampler(bs, &cmd->sampler, 1, 0);
                }
                else if(at == CMD_SET_TRK_PROP || at == CMD_SET_PARAM)
                {
                    cmd->prop_type = GetByte(bs);
                    GetSampler(bs, &cmd->sampler, 1, 0);
                }
                else if(at == CMD_SET_DFLT_TYPE)
                {
                    cmd->prop_type = GetByte(bs);
                }
                else if(at == CMD_SET_TEMPO)
                {
                    GetSampler(bs, &cmd->sampler, 1, 0);
                }
            }

            p->tracks[t->track_idx] = t;
        }
    }
}

static void Dispatch(Score* sc, Command* cmd, float rt)
{
    Track* t = &sc->tracks[sc->tidx];
    int tp = cmd->type & 0xf, u = cmd->type >> 4;

    if(tp == CMD_NOTE_OFF)
    {
        if(t->active_inst > -1)
            Trigger(sc, &sc->instruments[t->active_inst], TC_NOTE_OFF, rt - t->st);
    }
    else if(tp == CMD_NOTE_ON)
    {
        t->active_inst = cmd->inst;
        t->off = 0;
        t->st = rt;
        t->pitch = 440.0f * powf(2.0f, (cmd->pitch - 9.0f) / 12.0f);
        Trigger(sc, &sc->instruments[cmd->inst], TC_NOTE_ON, t->pitch);
    }
    else if(tp == CMD_SET_OSC_TYPE)
    {
        cmd->osc->type = u;
    }
    else if(tp == CMD_SET_FLT_TYPE)
    {
        cmd->filter->type = u;
    }
    else if(tp == CMD_SET_OSC_PROP)
    {
        (&cmd->osc->freq)[u] = cmd->sampler;
    }
    else if(tp == CMD_SET_FLT_PROP)
    {
        (&cmd->filter->gain)[u] = cmd->sampler;
    }
    else if(tp == CMD_SET_TRK_PROP)
    {
        if(cmd->prop_type >= TRK_PROP_D_GAIN)
            (&t->filter[u]->gain)[cmd->prop_type - TRK_PROP_D_GAIN] = cmd->sampler;
        else if(cmd->prop_type < TRK_PROP_DELAY)
            (&t->vol)[cmd->prop_type] = cmd->sampler;
        else if(cmd->prop_type == TRK_PROP_DELAY)
            t->delay[u] = cmd->sampler;
        else
            t->feedback[u] = cmd->sampler;
    }
    else if(tp == CMD_SET_DFLT_TYPE)
    {
        t->filter[u]->type = cmd->prop_type;
    }
    else if(tp == CMD_HALT)
    {
        t->active_inst = -1;
    }
    else if(tp == CMD_SET_PARAM)
    {
        t->params[cmd->prop_type] = cmd->sampler;
    }
    else if(tp == CMD_SET_TEMPO)
    {
        sc->tempo = cmd->sampler;
    }
    else // CMD_OFF
    {
        t->off = 1;
    }
}

static float Delay(Score* sc, Track* t, float rt, float v)
{
    float fv = 0.0f, dv = 0.0f, norm = 0.0f;

    for(int i = 0; i < 16; i++)
    {
        const float d = Sample(sc, &t->delay[i], rt);

        if(d == 0.0f)
            continue;
        
        const float d_ofs = SgnMod(d * 44100.0f, t->d_size - 1);
        float r_ofs = t->d_idx - d_ofs;

        if(r_ofs < 0.0f)
            r_ofs += t->d_size - 1;
        
        const unsigned int ro = (int)r_ofs;
        const float w = r_ofs - (float)ro;

        dv += (w * (t->d_buf[ro] - t->ly[i])) + t->d_buf[ro > 0 ? ro - 1 : t->d_size - 1],
        fv += RunFilter(sc, t->filter[i], rt, v + (Sample(sc, &t->feedback[i], rt) * dv));

        t->ly[i] = dv;
        norm += 1.0f;
    }

    if(norm > 0.0)
        norm = 1.0f / norm;
    
    t->d_buf[t->d_idx] = fv * norm;
    t->d_idx = t->d_idx < t->d_size - 1 ? t->d_idx + 1 : 0;

    return (v * Sample(sc, &t->dry, rt)) + (dv * norm);
}

void ProcessRow(Score* s, CmdList* cl, Pattern* cp, unsigned int ti, unsigned int cr)
{
    do
    {
        CmdBuf* cb = cp->tracks[ti];

        if(cb != NULL)
        {
            for(int i = cb->cmd_count - 1; i > -1; i--)
            {
                Command* c = cb->commands + i;

                if(c->row > cr)
                    continue;
                else if(c->row < cr)
                    break;
                
                if(cl->cidx < 0)
                    return;
                
                cl->commands[cl->cidx--] = c;
            }
        }

        if(cp->inherits != 0xffff && !(cb != NULL && cb->replace))
            ProcessRow(s, cl, &s->patterns[cp->inherits], ti, cr);
        
        ti = cb != NULL ? cb->inherits : 0xff;
    } while(ti != 0xff);
}

static AudioChunk* CreateAudioChunk()
{
    AudioChunk* ac = malloc(sizeof(AudioChunk));

    ac->idx = 0;
    ac->data = malloc(AC_SIZE * 2);
    ac->next = NULL;

    return ac;
}

static float GetFramesPerRow(Score* s, float pt)
{
    float tempo = Sample(s, &s->tempo, pt);

    tempo = tempo < 1.0f ? 1.0f : tempo > 512.0f ? 512.0f : tempo;
    
    return 44100.0f / (tempo / 15.0f); // (120 / 2) / 4
}

AudioBuffer* Mother(const unsigned char* bs, void (*pg_cb)(float))
{
    Score* s = malloc(sizeof(Score));

    Init(s, &bs);

    // Calc total duration
    unsigned int total_rows = 0, ar = 0;

    for(unsigned int i = 0; i < s->order_count; i++)
        total_rows += s->patterns[s->order[i]].row_count;

    float frames_acc = GetFramesPerRow(s, 0.0f), frames_per_row = 1.0f;
    int co = 0, cr = -1;
    Pattern* cp = &s->patterns[s->order[co]];
    CmdList cl;

    AudioBuffer* ab = malloc(sizeof(AudioBuffer));

    ab->size = 0;
    ab->score = s;

#if !defined(FRAMEDUMP) && !defined(NOMUSIC)
    AudioChunk* ac_root = CreateAudioChunk(), *ac = ac_root;
    unsigned int fr = 0;

    while(1)
    {
        float abst = (float)fr / 44100.0f;
        
        s->pt = cr > -1 ? (cr + (frames_acc / frames_per_row)) * 0.1f : 0.0f;
        frames_per_row = GetFramesPerRow(s, s->pt);
        s->abst = abst;

        if(frames_acc >= frames_per_row)
        {
            if(pg_cb)
                pg_cb((float)ar / (float)total_rows);

            cp->row_frames[cr > -1 ? cr : 0] = fr;
            frames_acc -= frames_per_row;

            cr++;
            ar++;

            if(cr >= (int)cp->row_count)
            {
                s->pt = 0.0f;
                cr = 0;
                co++;

                if(co < (int)s->order_count)
                    cp = &s->patterns[s->order[co]];
                else
                    break;
            }

            for(unsigned int i = 0; i < s->track_count; i++)
            {
                cl.cidx = 255;

                ProcessRow(s, &cl, cp, i, cr);

                if(cl.cidx < 255)
                {
                    s->tidx = i;

                    for(int ci = cl.cidx + 1; ci < 256; ci++)
                        Dispatch(s, cl.commands[ci], abst);
                }
            }
        }

        float l = 0.0f, r = 0.0f;

        for(unsigned int i = 0; i < s->track_count; i++)
        {
            Track* t = &s->tracks[i];

            if(t->off)
                continue;
            
            s->tidx = i;
            s->detune = Sample(s, &t->detune, s->pt);

            float iv = (t->active_inst > -1) ? Sample(s, &s->instruments[t->active_inst], abst - t->st) : 0.0f;
            CmdBuf* cb = cp->tracks[i];

            if(cb != NULL)
            {
                for(unsigned int j = 0; j < cb->r_count; j++)
                    iv += s->t_samps[cb->r_elems[j].tidx] * Sample(s, &cb->r_elems[j].value, s->pt);
            }
            
            const float vol = Sample(s, &t->vol, s->pt) * N3DB,
                        pan = Sample(s, &t->pan, s->pt) * 0.5f * PI,
                        v = t->d_size < 1 ? iv : Delay(s, t, s->pt, iv);

            s->t_samps[i] = v;

            l += v * vol * cosf(pan);
            r += v * vol * sinf(pan);
        }

        if(ac->idx < AC_SIZE)
        {
            ac->data[ac->idx] = CLIP_SCALE(l);
            ac->data[ac->idx+1] = CLIP_SCALE(r);
            ac->idx += 2;
            ab->size += 2;
        }
        else
        {
            AudioChunk* nac = CreateAudioChunk();

            ac->next = nac;
            ac = nac;
        }

        frames_acc += 1.0f;
        fr++;
    }

    ab->data = malloc(ab->size * sizeof(unsigned short));
    ac = ac_root;
    ar = 0;

    while(ac != NULL)
    {
        memcpy(ab->data + ar, ac->data, ac->idx * sizeof(unsigned short));
        ar += ac->idx;
        ac = ac->next;
    }
#endif

    return ab;
}